# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from dataclasses import asdict
from rich import print_json
from typer import Argument, Typer

from ..api import User
from .auth import get_access_key

app = Typer(no_args_is_help=True)

@app.command(name="retrieve", help="Retrieve a user.")
def retrieve_user (
    username: str=Argument(..., help="Username.")
) -> None:
    user = User.retrieve(username, access_key=get_access_key())
    user = asdict(user) if user else None
    print_json(data=user)


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from enum import Enum
from io import BytesIO
from mimetypes import guess_type
from pathlib import Path
from requests import put
from rich.progress import open as open_progress
from typing import Union

from .api import query

class UploadType (str, Enum):
    """
    Upload URL type.
    """
    Demo = "DEMO"
    Feature = "FEATURE"
    Graph = "GRAPH"
    Media = "MEDIA"
    Notebook = "NOTEBOOK"

class Storage:
    """
    Upload and download files.
    """

    @classmethod
    def create_upload_url (
        cls,
        name: str,
        type: UploadType,
        key: str=None
    ) -> str:
        """
        Create an upload URL.

        Parameters:
            name (str): File name.
            type (UploadType): Upload type.
            key (str): File key. This is useful for grouping related files.

        Returns:
            str: File upload URL.
        """
        # Query
        response = query(f"""
            mutation ($input: CreateUploadURLInput!) {{
                createUploadURL (input: $input)
            }}
            """,
            { "input": { "type": type, "name": name, "key": key } }
        )
        url = response["createUploadURL"]
        # Return
        return url

    @classmethod
    def upload ( # INCOMPLETE # `bytes` and `BytesIO` support` # Data URL limit
        cls,
        file: Union[str, Path, BytesIO, bytes],
        type: UploadType,
        name: str=None,
        key: str=None,
        data_url_limit: int=0,
        check_extension: bool=True,
        verbose: bool=False
    ) -> str:
        """
        Upload a file and return the URL.

        Parameters:
            file (str | Path | BytesIO | bytes): File path.
            type (UploadType): File type.
            name (str): File name. This MUST be provided if `file` is not a file path.
            key (str): File key. This is useful for grouping related files.
            data_url_limit (int): Return a data URL if the output feature is smaller than this limit (in bytes).
            check_extension (bool): Validate file extensions before uploading.
            verbose (bool): Print a progress bar for the upload.

        Returns:
            str: Upload URL.
        """
        EXTENSIONS = {
            UploadType.Demo: [".data", ".js", ".wasm"],
            UploadType.Feature: [],
            UploadType.Graph: [".mlmodel", ".onnx", ".tflite", ".pt", ".pth"],
            UploadType.Media: [".jpg", ".jpeg", ".png", ".gif"],
            UploadType.Notebook: [".ipynb"],
        }
        # Create path
        file = Path(file) if isinstance(file, str) else file
        # Check path
        if not file.exists():
            raise RuntimeError(f"Cannot upload {file.name} because the file does not exist")
        # Check file
        if not file.is_file():
            raise RuntimeError(f"Cannot upload {file.name} becaause it does not point to a file")
        # Check extension
        if check_extension and file.suffix not in EXTENSIONS[type]:
            raise RuntimeError(f"Cannot upload {file.name} because it is not a valid {type.name.lower()} file")
        # Get upload URL
        mime = guess_type(file, strict=False)[0] or "application/octet-stream"
        url = cls.create_upload_url(file.name, type, key=key)
        with open_progress(file, "rb", description=file.name, disable=not verbose) as f:
            put(url, data=f, headers={ "Content-Type": mime }).raise_for_status()
        # Return
        return url


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from dataclasses import asdict, dataclass
from io import BytesIO
from numpy import frombuffer
from PIL import Image
from requests import get
from typing import Any, Dict, List, Union
from urllib.request import urlopen

from ..api import query
from ..dtype import Dtype
from ..feature import Feature
from .feature import FeatureInput

@dataclass
class EndpointPrediction:
    """
    Endpoint prediction.

    Members:
        id (str): Prediction ID.
        tag (str): Endpoint tag.
        created (str): Date created.
        results (list): Prediction results.
        latency (float): Prediction latency in milliseconds.
        error (str): Prediction error. This is `null` if the prediction completed successfully.
        logs (str): Prediction logs.
    """
    id: str
    tag: str
    created: str
    results: List[Feature]
    latency: float
    error: str
    logs: str

    @classmethod
    def create (
        cls,
        tag: str,
        *features: List[FeatureInput],
        raw_outputs: bool=False,
        access_key: str=None,
        **inputs: Dict[str, Any],
    ) -> EndpointPrediction:
        """
        Create an endpoint prediction.

        Parameters:
            tag (str): Endpoint tag.
            features (list): Input features.
            raw_outputs (bool): Skip parsing output features into Pythonic data types.
            access_key (str): NatML access key.
            inputs (dict): Input features.

        Returns:
            EndpointPrediction: Endpoint prediction.
        """
        # Collect input features
        input_features = list(features) + [FeatureInput.from_value(value, name) for name, value in inputs.items()]
        input_features = [asdict(feature) for feature in input_features]
        # Query
        parsed_fields = "" if raw_outputs else "\n".join(_FEATURE_KEYS)
        response = query(f"""
            mutation ($input: CreateEndpointPredictionInput!) {{
                createEndpointPrediction (input: $input) {{
                    id
                    tag
                    created
                    results {{
                        data
                        type
                        shape
                        {parsed_fields}
                    }}
                    latency
                    error
                    logs
                }}
            }}""",
            { "input": { "tag": tag, "inputs": input_features, "client": "python" } },
            access_key=access_key
        )
        # Check prediction
        prediction = response["createEndpointPrediction"]
        if not prediction:
            return None
        # Parse outputs
        prediction["results"] = [_parse_output_feature(feature) for feature in prediction["results"]] if prediction["results"] and not raw_outputs else prediction["results"]
        prediction = EndpointPrediction(**prediction)
        # Return
        return prediction

def _parse_output_feature (feature: dict) -> Union[Feature, str, float, int, bool, Image.Image, list, dict]:
    data, type, shape = feature["data"], feature["type"], feature["shape"]
    # Handle image
    if type == Dtype.image:
        return Image.open(_download_feature_data(data))
    # Handle non-numeric scalars
    values = [feature.get(key, None) for key in _FEATURE_KEYS]
    scalar = next((value for value in values if value is not None), None)
    if scalar is not None:
        return scalar
    # Handle ndarray
    ARRAY_TYPES = [
        Dtype.int8, Dtype.int16, Dtype.int32, Dtype.int64,
        Dtype.uint8, Dtype.uint16, Dtype.uint32, Dtype.uint64,
        Dtype.float16, Dtype.float32, Dtype.float64, Dtype.bool
    ]
    if type in ARRAY_TYPES:
        # Create array
        array = frombuffer(_download_feature_data(data).getbuffer(), dtype=type).reshape(shape)
        return array if len(shape) > 0 else array.item()
    # Handle generic feature
    feature = Feature(**feature)
    return feature

def _download_feature_data (url: str) -> BytesIO:
    # Check if data URL
    if url.startswith("data:"):
        with urlopen(url) as response:
            return BytesIO(response.read())
    # Download
    response = get(url)
    result = BytesIO(response.content)
    return result

_FEATURE_KEYS = ["stringValue", "listValue", "dictValue"]


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from os import environ

# Define API URL and access key
api_url = "https://staging.api.natml.ai/graph" if environ.get("NATML_STAGING", None) else "https://api.natml.ai/graph"
access_key: str = environ.get("NATML_ACCESS_KEY", None)

# Import everything
from .api import *
from .version import __version__


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from dataclasses import dataclass

@dataclass
class GraphPrediction: # INCOMPLETE
    """
    Graph prediction.

    Members:
        id (str): Prediction ID.
        tag (str): Graph tag.
        graph (str): Prediction graph URL.
        code (str): Prediction code URL.
        created (str): Date created.
    """
    id: str
    tag: str
    graph: str
    code: str
    created: str

    @classmethod
    def create (
        cls,
        tag: str,
        device: str=None,
        access_key: str=None
    ) -> GraphPrediction:
        """
        Create a graph prediction.

        Parameters:
            tag (str): Graph tag.
            device (str): Device model identifier. This is used to optimize the graph prediction based on known hardware optimizations.
            access_key (str): NatML access key.

        Returns:
            GraphPrediction: Graph prediction.
        """
        pass


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from dataclasses import dataclass
from typing import Optional

@dataclass
class Tag:
    """
    Predictor tag.

    Members:
        username (str): Username.
        name (str): Name.
        variant (str): Variant.
    """
    username: str
    name: str
    variant: Optional[str] = None

def parse_tag (tag: str) -> Tag:
    """
    Parse a predictor tag.

    Parameters:
        tag (str): Tag string.

    Returns:
        Tag: Parsed tag.
    """
    username, full_name = tag.lower()[1:].split("/")
    name_components = full_name.split("@")
    name = name_components[0]
    variant = name_components[1] if len(name_components) > 1 else None
    result = Tag(username=username, name=name, variant=variant)
    return result

def serialize_tag (tag: Tag) -> str:
    """
    Serialize a predictor tag.

    Parameters:
        tag (Tag): Tag.

    Returns:
        str: Serialized tag.
    """
    username, name, variant = tag
    suffix = f"@{variant}" if variant else ""
    result = f"@{username}/{name}{suffix}"
    return result


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union

from .api import query
from .storage import Storage, UploadType
from .tag import Tag, parse_tag, serialize_tag

@dataclass
class Graph:
    """
    Predictor graph.

    Members:
        variant (str): Graph variant.
        format (GraphFormat): Graph format.
        status (GraphStatus): Graph status.
        encrypted (bool): Whether the graph is encrypted.
        created (str): Date created.
        error (str): Graph provisioning error. This is populated when the graph `status` is `INVALID`.
    """
    variant: str
    format: GraphFormat
    status: GraphStatus
    encrypted: bool
    created: str
    error: Optional[str] = None
    FIELDS = f"""
    variant
    format
    status
    encrypted
    created
    error
    """

    @classmethod
    def retrieve (
        cls,
        tag: str,
        format: GraphFormat,
        access_key: str=None
    ) -> Graph:
        """
        Retrieve a predictor graph.

        Parameters:
            tag (str): Graph tag. If the tag does not contain a variant then the variant defaults to `main`.
            format (GraphFormat): Graph format.
            access_key (str): NatML access key.

        Returns:
            Graph: Predictor graph.
        """
        # Ensure this is a predictor tag
        tag = parse_tag(tag)
        variant = tag.variant or "main"
        tag = Tag(tag.username, tag.name)
        tag = serialize_tag(tag)
        # There isn't a query to get a specific graph so just filter for now
        graphs = cls.list(tag, access_key=access_key)
        graph = next((x for x in graphs if x.variant == variant and x.format == format), None) if graphs else None
        # Return
        return graph

    @classmethod
    def list (
        cls,
        tag: str,
        access_key: str=None
    ) -> List[Graph]:
        """
        Retrieve all predictor graphs.

        Parameters:
            tag (str): Predictor tag. This MUST NOT be a variant tag.
            access_key (str): NatML access key.

        Returns:
            list: Predictor graphs.
        """
        # Request
        response = query(f"""
            query ($input: PredictorInput!) {{
                predictor (input: $input) {{
                    graphs {{
                        {Graph.FIELDS}
                    }}
                }}
            }}""",
            { "input": { "tag": tag } },
            access_key=access_key
        )
        # Check predictor
        predictor = response["predictor"]
        if not predictor:
            return None
        # Get graphs
        graphs = [Graph(**graph) for graph in predictor["graphs"]]
        return graphs

    @classmethod
    def create (
        cls,
        tag: str,
        graph: Union[str, Path],
        format: GraphFormat,
        access_key: str=None
    ) -> Graph:
        """
        Create a predictor graph.

        Parameters:
            tag (str): Graph tag. If the tag does not contain a variant then the variant defaults to `main`.
            graph (str | Path): Source graph URL or path.
            format (GraphFormat): Target graph format.
            access_key (str): NatML access key.

        Returns:
            Graph: Created graph.
        """
        # Upload graph
        url = Storage.upload(graph, UploadType.Graph, check_extension=True) if isinstance(graph, Path) else graph
        # Query
        response = query(f"""
            mutation ($input: CreateGraphInput!) {{
                createGraph (input: $input) {{
                    {Graph.FIELDS}
                }}
            }}""",
            { "input": { "tag": tag, "graph": url, "format": format } },
            access_key=access_key
        )
        # Create endpoint
        graph = response["createGraph"]
        graph = Graph(**graph)
        # Return
        return graph

    @classmethod
    def delete (
        cls,
        tag: str,
        format: GraphFormat,
        access_key: str=None
    ) -> bool:
        """
        Delete a predictor graph.

        Parameters:
            tag (str): Graph tag. If the tag does not contain a variant then the variant defaults to `main`.
            format (GraphFormat): Graph format.
            access_key (str): NatML access key.

        Returns:
            bool: Whether the graph was successfully deleted.
        """
        # Query
        response = query(f"""
            mutation ($input: DeleteGraphInput!) {{
                deleteGraph (input: $input)
            }}""",
            { "input": { "tag": tag, "format": format } },
            access_key=access_key
        )
        # Return
        result = response["deleteGraph"]
        return result
    
class GraphFormat (str, Enum):
    """
    Graph format.
    """
    CoreML = "COREML"
    ONNX = "ONNX"
    TFLite = "TFLITE"

class GraphStatus (str, Enum):
    """
    Graph status.
    """
    Provisioning = "PROVISIONING"
    Active = "ACTIVE"
    Invalid = "INVALID"


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from dataclasses import dataclass

from .api import query
from .profile import Profile

@dataclass
class User (Profile):
    """
    User.
    """

    @classmethod
    def retrieve (
        cls,
        username: str=None,
        access_key: str=None
    ) -> User:
        """
        Retrieve a user.

        Parameters:
            username (str): Username. If `None`, this will retrieve the currently authenticated user.
            access_key (str): NatML access key.

        Returns:
            User: User.
        """
        # Query
        response = query(f"""
            query {"($input: UserInput)" if username else ""} {{
                user {"(input: $input)" if username else ""} {{
                    {Profile.FIELDS}
                    {"email" if not username else ""}
                }}
            }}
            """,
            { "input": { "username": username } },
            access_key=access_key
        )
        # Create user
        user = response["user"]
        user = User(**user) if user else None
        # Return
        return user


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from enum import Enum

class License (str, Enum):
    """
    Predictor license.
    """
    OpenRAIL = "openrail"
    Apache20 = "apache-2.0"
    BSD = "bsd"
    BSD2Clause = "bsd-2-clause"
    BSD3Clause = "bsd-3-clause"
    AGPLv3 = "agpl-3.0"
    GPL = "gpl"
    GPLv2 = "gpl-2.0"
    GPLv3 = "gpl-3.0"
    LGPL = "lgpl"
    LGPLv21 = "lgpl-2.1"
    LGPLv3 = "lgpl-3.0"
    ISC = "isc"
    MIT = "mit"
    OSLv3 = "osl-3.0"
    Unlicense = "unlicense"
    Unknown = "unknown"
    Other = "other" 
    WTFPL = "wtfpl"
    CreativeCommons = "cc"
    CreativeCommons0v10 = "cc0-1.0"
    CreativeCommonsv20 = "cc-by-2.0"
    CreativeCommonsv25 = "cc-by-2.5"
    CreativeCommonsv30 = "cc-by-3.0"
    CreativeCommonsv40 = "cc-by-4.0"
    CreativeCommonsShareAlikev30 = "cc-by-sa-3.0"
    CreativeCommonsShareAlike40 = "cc-by-sa-4.0"
    CreativeCommonsNonCommercialv20 = "cc-by-nc-2.0"
    CreativeCommonsNonCommercialv30 = "cc-by-nc-3.0"
    CreativeCommonsNonCommercialv40 = "cc-by-nc-4.0"
    CreativeCommonsNoDerivativesv40 = "cc-by-nd-4.0"
    CreativeCommonsNonCommercialNoDerivativesv30 = "cc-by-nc-nd-3.0"
    CreativeCommonsNonCommercialNoDerivativesv40 ="cc-by-nc-nd-4.0"
    CreativeCommonsNonCommercialShareAlikev20 = "cc-by-nc-sa-2.0"
    CreativeCommonsNonCommercialShareAlikev30 = "cc-by-nc-sa-3.0"
    CreativeCommonsNonCommercialShareAlikev40 = "cc-by-nc-sa-4.0"


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from dataclasses import dataclass
from typing import Optional

@dataclass
class Profile:
    """
    Entity with a public NatML profile.

    Members:
        username (str): Username.
        email (str): User email address.
        created (str): Date created.
        name (str): User display name.
        avatar (str): User avatar URL.
        bio (str): User bio.
        website (str): User website.
        github (str): User GitHub handle.
    """
    username: str
    email: Optional[str] = None
    created: Optional[str] = None
    name: Optional[str] = None
    avatar: Optional[str] = None
    bio: Optional[str] = None
    website: Optional[str] = None
    github: Optional[str] = None
    FIELDS = f"""
    username
    created
    name
    avatar
    bio
    website
    github
    """


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from rich import print
from typer import Exit, Option
from typing import Callable
from webbrowser import open as open_browser

from ..version import __version__

def create_learn_callback (url: str) -> Callable[[bool], None]:
    # Define callback
    def callback (value: bool):
        if value:
            open_browser(url)
            raise Exit()
    # Return
    return callback

def _version_callback (value: bool):
    if value:
        print(__version__)
        raise Exit()

def cli_options (
    learn: bool = Option(None, "--learn", callback=create_learn_callback("https://docs.natml.ai"), help="Learn about NatML."),
    version: bool = Option(None, "--version", callback=_version_callback, help="Get the NatML CLI version.")
):
    pass


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from requests import post

import natml

def query (query: str, variables: dict=None, access_key: str=None) -> dict:
    """
    Issue a query to the NatML Graph API.

    Parameters:
        query (str): Graph query.
        variables (dict): Input variables.
        access_key (str): NatML access key.

    Returns:
        dict: Response dictionary.
    """
    access_key = access_key or natml.access_key
    headers = { "Authorization": f"Bearer {access_key}" } if access_key else { }
    response = post(
        natml.api_url,
        json={ "query": query, "variables": variables },
        headers=headers
    ).json()
    # Check error
    if "errors" in response:
        raise RuntimeError(response["errors"][0]["message"])
    # Return
    result = response["data"]
    return result


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from dataclasses import asdict
from pathlib import Path
from rich import print_json
from typer import Argument, Context, Option, Typer

from ..api import Graph, GraphFormat
from .auth import get_access_key
from .misc import create_learn_callback

app = Typer(no_args_is_help=True)

@app.command(name="retrieve", help="Retrieve a predictor graph.")
def retrieve_graph (
    tag: str=Argument(..., help="Graph tag. If the tag does not contain a variant then the variant defaults to `main`."),
    format: GraphFormat=Argument(..., case_sensitive=False, help="Graph format.")
) -> None:
    graph = Graph.retrieve(tag, format, access_key=get_access_key())
    graph = asdict(graph) if graph else None
    print_json(data=graph)

@app.command(name="list", help="List all predictor graphs.")
def list_graph (
    tag: str=Argument(..., help="Predictor tag. This MUST NOT be a variant tag.")
) -> None:
    graphs = Graph.list(tag, access_key=get_access_key())
    graphs = [asdict(graph) for graph in graphs] if graphs else None
    print_json(data=graphs)

@app.command(name="create", help="Create a predictor graph.")
def create_graph (
    tag: str=Argument(..., help="Graph tag. If the tag does not contain a variant then the variant defaults to `main`."),
    graph: Path=Argument(..., help="Path to ML graph."),
    format: GraphFormat=Argument(..., case_sensitive=False, help="Target graph format.")
) -> None:
    graph = Graph.create(tag, graph, format, access_key=get_access_key())
    graph = asdict(graph)
    print_json(data=graph)

@app.command(name="delete", help="Delete a predictor graph.")
def delete_graph (
    tag: str=Argument(..., help="Graph tag. If the tag does not contain a variant then the variant defaults to `main`."),
    format: GraphFormat=Argument(..., case_sensitive=False, help="Graph format.")
) -> None:
    result = Graph.delete(tag, format, access_key=get_access_key())
    print_json(data=result)

@app.command(
    name="predict",
    context_settings={ "allow_extra_args": True, "ignore_unknown_options": True },
    help="Make a prediction with a predictor graph."
)
def predict (
    tag: str = Argument(..., help="Graph tag."),
    raw_outputs: bool = Option(False, "--raw-outputs", help="Generate raw output features instead of parsing."),
    context: Context = 0
) -> None:
    raise RuntimeError("Making edge predictions is not yet supported in NatML Python SDK")

@app.callback()
def graph_options (
    learn: bool = Option(None, "--learn", callback=create_learn_callback("https://docs.natml.ai/graph/graphs/type"), help="Learn about predictor graphs in NatML.")
):
    pass


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from base64 import b64encode
from dataclasses import dataclass
from io import BytesIO
from numpy import ndarray
from PIL import Image
from typing import Any, Dict, List, Optional, Union

from ..dtype import Dtype

@dataclass
class FeatureInput:
    """
    Prediction input feature.

    Members:
        name (str): Feature name. This MUST match the input parameter name defined by the predictor endpoint.
        data (str): Feature data URL. This can be a web URL or a data URL.
        type (Dtype): Feature data type.
        shape (list): Feature shape. This MUST be provided for array features.
    """
    name: str
    data: str = None
    type: Dtype = None
    shape: Optional[List[int]] = None
    stringValue: str = None
    floatValue: float = None
    floatArray: List[float] = None
    intValue: int = None
    intArray: List[int] = None
    boolValue: bool = None
    listValue: list = None
    dictValue: Dict[str, Any] = None

    @classmethod
    def from_value (
        cls,
        value: Union[ndarray, str, float, int, bool, List, Dict[str, any], Image.Image],
        name: str
    ) -> FeatureInput:
        """
        Create a feature input from a given value.
        
        Parameters:
            value (any): Value.
            name (str): Feature name.

        Returns:
            FeatureInput: Feature input.
        """
        # Array
        if isinstance(value, ndarray):
            encoded_data = b64encode(value).decode("ascii")
            data = f"data:application/octet-stream;base64,{encoded_data}"
            return FeatureInput(name, data, value.dtype.name, list(value.shape))
        # String
        if isinstance(value, str):
            return FeatureInput(name, stringValue=value)
        # Float
        if isinstance(value, float):
            return FeatureInput(name, floatValue=value)
        # Boolean
        if isinstance(value, bool):
            return FeatureInput(name, boolValue=value)
        # Integer
        if isinstance(value, int):
            return FeatureInput(name, intValue=value)
        # List
        if isinstance(value, list):
            return FeatureInput(name, listValue=value)
        # Dict
        if isinstance(value, dict):
            return FeatureInput(name, dictValue=value)
        # Image
        if isinstance(value, Image.Image):
            image_buffer = BytesIO()
            channels = { "L": 1, "RGB": 3, "RGBA": 4 }[value.mode]
            format = "PNG" if value.mode == "RGBA" else "JPEG"
            value.save(image_buffer, format=format)
            encoded_data = b64encode(image_buffer.getvalue()).decode("ascii")
            data = f"data:{value.get_format_mimetype()};base64,{encoded_data}"
            shape = [1, value.height, value.width, channels]
            return FeatureInput(name, data, Dtype.image, shape)
        # Unsupported
        raise RuntimeError(f"Cannot create input feature for value {value} of type {type(value)}")


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from typer import Argument, Context, Option, Typer

from .endpoints import predict as predict_endpoint
from .graphs import predict as predict_graph

def predict ( # INCOMPLETE
    tag: str = Argument(..., help="Predictor tag."),
    raw_outputs: bool = Option(False, "--raw-outputs", help="Generate raw output features instead of parsing."),
    context: Context = 0
) -> None:
    predict_endpoint(tag, raw_outputs, context)


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from dataclasses import asdict
from rich import print_json
from typer import Argument, Option, Typer
from typing import List, Tuple

from ..api import AccessMode, License, Predictor, PredictorStatus
from .auth import get_access_key
from .misc import create_learn_callback

app = Typer(no_args_is_help=True)

@app.command(name="retrieve", help="Retrieve a predictor.")
def retrieve_predictor (
    tag: str=Argument(..., help="Predictor tag.")
) -> None:
    predictor = Predictor.retrieve(tag, access_key=get_access_key())
    predictor = asdict(predictor) if predictor else None
    print_json(data=predictor)

@app.command(name="list", help="List available predictors.")
def list_predictors (
    mine: bool=Option(False, "--mine", help="Fetch only predictors owned by me."),
    status: PredictorStatus=Option(None, case_sensitive=False, help="Predictor status. This only applies when `--mine` is specified."),
    offset: int=Option(None, help="Pagination offset."),
    count: int=Option(None, help="Pagination count.")
) -> None:
    predictors = Predictor.list(mine=mine, status=status, offset=offset, count=count, access_key=get_access_key())
    predictors = [asdict(predictor) for predictor in predictors]
    print_json(data=predictors)

@app.command(name="search", help="Search predictors.")
def search_predictors (
    query: str=Argument(..., help="Search query."),
    offset: int=Option(None, help="Pagination offset."),
    count: int=Option(None, help="Pagination count.")
) -> None:
    predictors = Predictor.search(query=query, offset=offset, count=count, access_key=get_access_key())
    predictors = [asdict(predictor) for predictor in predictors]
    print_json(data=predictors)

@app.command(name="create", help="Create a predictor.")
def create_predictor (
    tag: str=Argument(..., help="Predictor tag."),
    description: str=Option("", help="Predictor description."),
    access: AccessMode=Option(AccessMode.Private, case_sensitive=False, help="Predictor access mode.")
) -> None:
    predictor = Predictor.create(tag, description=description, access=access, access_key=get_access_key())
    print_json(data=asdict(predictor))

@app.command(name="update", help="Update a predictor.")
def update_predictor (
    tag: str=Argument(..., help="Predictor tag."),
    description: str=Option(None, help="Predictor description."),
    access: AccessMode=Option(None, case_sensitive=False, help="Predictor access mode."),
    license: License=Option(None, case_sensitive=False, help="Predictor license."),
    topics: List[str]=Option(None, help="Predictor topics."),
    media: str=Option(None, help="Predictor media."),
    delete_media: bool=Option(False, "--delete-media", help="Delete existing media."),
) -> None:
    predictor = Predictor.update(
        tag,
        description=description,
        access=access,
        license=license,
        topics=topics,
        media=media,
        delete_media=delete_media,
        access_key=get_access_key()
    )
    print_json(data=asdict(predictor))

@app.command(name="delete", help="Delete a draft predictor.")
def delete_predictor (
    tag: str=Argument(..., help="Predictor tag.")
) -> None:
    result = Predictor.delete(tag, access_key=get_access_key())
    print_json(data=result)

@app.command(name="review", help="Review a draft predictor for any issues that might prevent it from being published.")
def review_predictor (
    tag: str=Argument(..., help="Predictor tag.")
) -> None:
    result = Predictor.review(tag, access_key=get_access_key())
    print_json(data=result)

@app.command(name="publish", help="Publish a draft predictor.")
def publish_predictor (
    tag: str=Argument(..., help="Predictor tag.")
) -> None:
    predictor = Predictor.publish(tag, access_key=get_access_key())
    print_json(data=asdict(predictor))

@app.command(name="archive", help="Archive a published predictor.")
def archive_predictor (
    tag: str=Argument(..., help="Predictor tag.")
) -> None:
    predictor = Predictor.archive(tag, access_key=get_access_key())
    print_json(data=asdict(predictor))

@app.callback()
def predictor_options (
    learn: bool = Option(None, "--learn", callback=create_learn_callback("https://docs.natml.ai/graph/predictors/type"), help="Learn about predictors in NatML.")
):
    pass


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from typer import Typer

from .auth import app as auth_app
from .demos import app as demos_app
from .endpoints import app as endpoints_app
from .graphs import app as graphs_app
from .misc import cli_options
from .predict import predict
from .predictors import app as predictors_app
from .users import app as users_app
from ..version import __version__

# Define CLI
app = Typer(
    name=f"NatML CLI {__version__}",
    no_args_is_help=True,
    pretty_exceptions_show_locals=False,
    pretty_exceptions_short=True,
)

# Add top level options
app.callback()(cli_options)

# Add subcommands
app.add_typer(auth_app, name="auth", help="Login, logout, and perform other authentication tasks.")
app.add_typer(predictors_app, name="predictors", help="Manage predictors.")
app.add_typer(graphs_app, name="graphs", help="Manage predictor graphs.")
app.add_typer(endpoints_app, name="endpoints", help="Manage predictor endpoints.")
app.add_typer(users_app, name="users", help="Manage users.")
app.add_typer(demos_app, name="demos", help="Manage predictor demos.")

# Add top-level commands
app.command(
    name="predict",
    context_settings={ "allow_extra_args": True, "ignore_unknown_options": True },
    help="Make a prediction with a predictor."
)(predict)

# Run
if __name__ == "__main__":
    app()


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union

from .api import query
from .license import License
from .profile import Profile
from .storage import Storage, UploadType

class AccessMode (str, Enum):
    """
    Predictor access mode.
    """
    Public = "PUBLIC"
    Private = "PRIVATE"

class PredictorStatus (str, Enum):
    """
    Predictor status.
    """
    Draft = "DRAFT"
    Published = "PUBLISHED"
    Archived = "ARCHIVED"

@dataclass
class Predictor:
    """
    Predictor.

    Members:
        tag (str): Predictor tag.
        owner (Profile): Predictor owner.
        name (str): Predictor name.
        description (str): Predictor description.
        status (PredictorStatus): Predictor status.
        access (AccessMode): Predictor access.
        license (License): Predictor license.
        topics (list): Predictor topics.
        created (str): Date created.
        media (str): Predictor media.
    """
    tag: str
    owner: Profile
    name: str
    description: str
    status: PredictorStatus
    access: AccessMode
    license: License
    topics: Optional[List[str]] = None
    created: Optional[str] = None
    media: Optional[str] = None
    FIELDS = f"""
    tag
    owner {{
        {Profile.FIELDS}
    }}
    name
    description
    status
    access
    license
    topics
    created
    media
    """

    def __post_init__ (self):
        self.owner = Profile(**self.owner, email=None) if isinstance(self.owner, dict) else self.owner

    @classmethod
    def retrieve (
        cls,
        tag: str,
        access_key: str=None
    ) -> Predictor:
        """
        Retrieve a predictor.

        Parameters:
            tag (str): Predictor tag. This MUST NOT be a variant tag.
            access_key (str): NatML access key.

        Returns:
            Predictor: Predictor.
        """
        # Query
        response = query(f"""
            query ($input: PredictorInput!) {{
                predictor (input: $input) {{
                    {Predictor.FIELDS}
                }}
            }}
            """,
            { "input": { "tag": tag } },
            access_key=access_key
        )
        # Create predictor
        predictor = response["predictor"]
        predictor = Predictor(**predictor) if predictor else None
        # Return
        return predictor

    @classmethod
    def list (
        cls,
        mine: bool=False,
        status: PredictorStatus=None,
        offset: int=None,
        count: int=None,
        access_key: str=None
    ) -> List[Predictor]:
        """
        View available predictors.
        
        Parameters:
            mine (bool): Fetch only predictors owned by me.
            status (PredictorStatus): Predictor status. This only applies when `mine` is `True`.
            offset (int): Pagination offset.
            count (int): Pagination count.
            access_key (str): NatML access key.

        Returns:
            list: Predictors.
        """
        # Collect inputs
        input = { "mine": mine, "status": status, "offset": offset, "count": count }
        input = { k: v for k, v in input.items() if v is not None }
        # Query
        response = query(f"""
            query ($input: PredictorsInput!) {{
                predictors (input: $input) {{
                    {Predictor.FIELDS}
                }}
            }}
            """,
            { "input": input },
            access_key=access_key
        )
        # Create predictors
        predictors = response["predictors"]
        predictors = [Predictor(**predictor) for predictor in predictors]
        # Return
        return predictors

    @classmethod
    def search (
        cls,
        query: str,
        offset: int=None,
        count: int=None,
        access_key: str=None
    ) -> List[Predictor]:
        """
        Search predictors.

        Parameters:
            query (str): Search query.
            access_key (str): NatML access key.

        Returns:
            list: Relevant predictors.
        """
        # Query
        response = query(f"""
            query ($input: PredictorsInput!) {{
                predictors (input: $input) {{
                    {Predictor.FIELDS}
                }}
            }}
            """,
            { "input": { "query": query, "offset": offset, "count": count } },
            access_key=access_key
        )
        # Create predictors
        predictors = response["predictors"]
        predictors = [Predictor(**predictor) for predictor in predictors]
        # Return
        return predictors

    @classmethod
    def create (
        cls,
        tag: str,
        description: str=None,
        access: AccessMode=AccessMode.Private,
        access_key: str=None
    ) -> Predictor:
        """
        Create a predictor.

        Parameters:
            tag (str): Predictor tag.
            description (str): Predictor description. This supports Markdown.
            access (AccessMode): Predictor access mode.
            access_key (str): NatML access key.

        Returns:
            Predictor: Created predictor.
        """
        # Query
        response = query(f"""
            mutation ($input: CreatePredictorInput!) {{
                createPredictor (input: $input) {{
                    {Predictor.FIELDS}
                }}
            }}
            """,
            { "input": { "tag": tag, "description": description, "access": access } },
            access_key=access_key
        )
        # Create predictor
        predictor = response["createPredictor"]
        predictor = Predictor(**predictor) if predictor else None
        # Return
        return predictor

    @classmethod
    def update (
        cls,
        tag: str,
        description: str=None,
        access: AccessMode=None,
        license: License=None,
        topics: List[str]=None,
        media: Union[str,Path]=None,
        delete_media: bool=False,
        access_key: str=None
    ) -> Predictor:
        """
        Update a predictor.

        Parameters:
            tag (str): Predictor tag.
            description (str): Predictor description. This supports Markdown.
            access (AccessMode): Predictor access mode.
            license (License): Predictor license.
            topics (list): Predictor topics.
            media (str | Path): Predictor media URL or path. Pass an empty string to unset.
            delete_media (bool): Delete existing media.
            access_key (str): NatML access key.

        Returns:
            Predictor: Updated predictor.
        """
        # Upload media
        media = Storage.upload(media, UploadType.Media, check_extension=True) if isinstance(media, Path) else media
        # Gather inputs
        preserve_keys = [k for k, v in {
            "media": delete_media,
        }.items() if v]
        input = {
            "tag": tag,
            "description": description,
            "access": access,
            "license": license,
            "topics": topics,
            "media": media,
        }
        input = { k: v for k, v in input.items() if v is not None or k in preserve_keys }
        # Query
        response = query(f"""
            mutation ($input: UpdatePredictorInput!) {{
                updatePredictor (input: $input) {{
                    {Predictor.FIELDS}
                }}
            }}
            """,
            { "input": input },
            access_key=access_key
        )
        # Create predictor
        predictor = response["updatePredictor"]
        predictor = Predictor(**predictor) if predictor else None
        # Return
        return predictor

    @classmethod
    def delete (
        cls,
        tag: str,
        access_key: str=None
    ) -> bool:
        """
        Delete a draft predictor.

        Parameters:
            tag (str): Predictor tag.
            access_key (str): NatML access key.

        Returns:
            bool: Whether the predictor was successfully deleted.
        """
        # Query
        response = query(f"""
            mutation ($input: DeletePredictorInput!) {{
                deletePredictor (input: $input)
            }}
            """,
            { "input": { "tag": tag } },
            access_key=access_key
        )
        # Return
        result = response["deletePredictor"]
        return result

    @classmethod
    def publish (
        cls,
        tag: str,
        access_key: str=None
    ) -> Predictor:
        """
        Publish a draft predictor.

        Parameters:
            tag (str): Predictor tag.
            access_key (str): NatML access key.

        Returns:
            Predictor: Published predictor.
        """
        # Query
        response = query(f"""
            mutation ($input: PublishPredictorInput!) {{
                publishPredictor (input: $input) {{
                    {Predictor.FIELDS}
                }}
            }}
            """,
            { "input": { "tag": tag } },
            access_key=access_key
        )
        # Create predictor
        predictor = response["publishPredictor"]
        predictor = Predictor(**predictor) if predictor else None
        # Return
        return predictor

    @classmethod
    def archive (
        cls,
        tag: str,
        access_key: str=None
    ) -> Predictor:
        """
        Archive a published predictor.

        Parameters:
            tag (str): Predictor tag.
            access_key (str): NatML access key.

        Returns:
            Predictor: Archived predictor.
        """
        # Query
        response = query(f"""
            mutation ($input: ArchivePredictorInput!) {{
                archivePredictor (input: $input) {{
                    {Predictor.FIELDS}
                }}
            }}
            """,
            { "input": { "tag": tag } },
            access_key=access_key
        )
        # Create predictor
        predictor = response["archivePredictor"]
        predictor = Predictor(**predictor) if predictor else None
        # Return
        return predictor


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from .dtype import Dtype
from .endpoint import Endpoint, EndpointAcceleration, EndpointStatus, EndpointType, EndpointSignature, EndpointParameter
from .feature import Feature
from .graph import Graph, GraphFormat, GraphStatus
from .license import License
from .prediction import EndpointPrediction, FeatureInput
from .predictor import Predictor, PredictorStatus, AccessMode
from .profile import Profile
from .storage import Storage, UploadType
from .user import User


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from dataclasses import asdict
from filetype import image_match
from io import BytesIO
from numpy import ndarray
from pathlib import Path
from PIL import Image
from rich import print_json
from tempfile import mkstemp
from typer import Argument, Context, Option, Typer
from typing import List, Optional, Union
from typing_extensions import Annotated

from ..api import Endpoint, EndpointType, EndpointAcceleration, EndpointPrediction
from .auth import get_access_key
from .misc import create_learn_callback

app = Typer(no_args_is_help=True)

@app.command(name="retrieve", help="Retrieve a predictor endpoint.")
def retrieve_endpoint (
    tag: str=Argument(..., help="Endpoint tag. If the tag does not contain a variant then the variant defaults to `main`.")
) -> None:
    endpoint = Endpoint.retrieve(tag, access_key=get_access_key())
    endpoint = asdict(endpoint) if endpoint else None
    print_json(data=endpoint)

@app.command(name="list", help="List all predictor endpoints.")
def list_endpoints (
    tag: str=Argument(..., help="Predictor tag. This MUST NOT be a variant tag.")
) -> None:
    endpoints = Endpoint.list(tag, access_key=get_access_key())
    endpoints = [asdict(endpoint) for endpoint in endpoints] if endpoints else None
    print_json(data=endpoints)

@app.command(name="create", help="Create a predictor endpoint.")
def create_endpoint (
    tag: str=Argument(..., help="Endpoint tag."),
    notebook: Path=Argument(..., help="Path to endpoint notebook."),
    type: EndpointType=Option(EndpointType.Serverless, case_sensitive=False, help="Endpoint type."),
    acceleration: EndpointAcceleration=Option(EndpointAcceleration.CPU, case_sensitive=False, help="Endpoint acceleration."),
    environment: Annotated[Optional[List[str]], Option(default=[])] = None
) -> None:
    environment = { e.split("=")[0].strip(): e.split("=")[1].strip() for e in environment }
    endpoint = Endpoint.create(tag, notebook, type, acceleration, environment=environment, access_key=get_access_key())
    endpoint = asdict(endpoint)
    print_json(data=endpoint)

@app.command(name="delete", help="Delete a predictor endpoint.")
def delete_endpoint (
    tag: str=Argument(..., help="Endpoint tag.")
) -> None:
    result = Endpoint.delete(tag, access_key=get_access_key())
    print_json(data=result)

@app.command(
    name="predict",
    context_settings={ "allow_extra_args": True, "ignore_unknown_options": True },
    help="Make a prediction with a predictor endpoint."
)
def predict (
    tag: str = Argument(..., help="Endpoint tag."),
    raw_outputs: bool = Option(False, "--raw-outputs", help="Generate raw output features instead of parsing."),
    context: Context = 0
) -> None:
    inputs = { context.args[i].replace("-", ""): _parse_value(context.args[i+1]) for i in range(0, len(context.args), 2) }
    session = EndpointPrediction.create(tag, **inputs, raw_outputs=raw_outputs, access_key=get_access_key())
    images = [feature for feature in session.results if isinstance(feature, Image.Image)] if session.results is not None else []
    session.results = [_serialize_feature(feature) for feature in session.results] if session.results is not None else None
    print_json(data=asdict(session))
    for image in images:
        image.show()

@app.callback()
def endpoint_options (
    learn: bool = Option(None, "--learn", callback=create_learn_callback("https://docs.natml.ai/graph/endpoints/type"), help="Learn about predictor endpoints in NatML.")
):
    pass

def _parse_value (value: str):
    """
    Parse a value from a CLI argument.

    Parameters:
        value (str): CLI input argument.

    Returns:
        bool | int | float | str | Path: Parsed value.
    """
    # Boolean
    if value.lower() == "true":
        return True
    if value.lower() == "false":
        return False
    # Integer
    try:
        return int(value)
    except ValueError:
        pass
    # Float
    try:
        return float(value)
    except ValueError:
        pass
    # File
    if value.startswith("@"):
        return _parse_file(Path(value[1:]))
    # String
    return value

def _parse_file (path: Path) -> Union[Image.Image, BytesIO]:
    """
    Parse a file from a given path.

    Parameters:
        path (Path): File path.

    Returns:
        Image | BytesIO: File.
    """
    # Load image
    if image_match(path):
        return Image.open(path)
    # Load buffer
    with open(path, "rb") as f:
        return BytesIO(f.read())
    
def _serialize_feature (feature):
    # Convert ndarray to list
    if isinstance(feature, ndarray):
        return feature.tolist()
    # Write image
    if isinstance(feature, Image.Image):
        _, path = mkstemp(suffix=".png" if feature.mode == "RGBA" else ".jpg")
        feature.save(path)
        return path
    # Return    
    return feature


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from dataclasses import dataclass
from typing import List, Optional

from .dtype import Dtype

@dataclass
class Feature:
    """
    Prediction feature.

    Members:
        data (str): Feature data URL.
        type (Dtype): Feature data type.
        shape (list): Feature shape. This is `None` if shape information is not available or applicable.
    """
    data: str
    type: Dtype
    shape: Optional[List[int]] = None


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from setuptools import find_packages, setup

# Get readme
with open("README.md", "r") as readme:
    long_description = readme.read()

# Get version
with open("natml/version.py") as version_source:
    gvars = {}
    exec(version_source.read(), gvars)
    version = gvars["__version__"]

# Setup
setup(
    name="natml",
    version=version,
    author="NatML Inc.",
    author_email="hi@natml.ai",
    description="Zero deployment machine learning.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="Apache License 2.0",
	python_requires=">=3.7",
    install_requires=[
        "filetype",
        "numpy",
        "pillow",
        "requests",
        "rich",
        "typer"
    ],
    url="https://natml.ai",
    packages=find_packages(
        include=["natml", "natml.*"],
        exclude=["test", "examples"]
    ),
    entry_points={
        "console_scripts": [
            "natml=natml.cli.__init__:app"
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Image Recognition",
        "Topic :: Software Development :: Libraries",
    ],
    project_urls={
        "Documentation": "https://docs.natml.ai/python",
        "Source": "https://github.com/natmlx/natml-py"
    },
)


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from pathlib import Path
from re import search
from rich import print, print_json
from typer import Argument, Option, Typer
from uuid import uuid4
from webbrowser import open as open_browser

from ..api import Storage, UploadType
from ..api.api import query
from .auth import get_access_key
from .misc import create_learn_callback

app = Typer(no_args_is_help=True)

@app.command(name="create", help="Create a Unity WebGL demo.")
def create_demo (
    tag: str=Argument(..., help="Demo tag."),
    path: Path=Option(None, help="Unity WebGL build directory. This defaults to the current working directory.")
) -> None:
    path = path if path is not None else Path.cwd()
    # Check index
    index = path / "index.html"
    if not index.is_file():
        raise RuntimeError("Count not find WebGL index.html")
    # Get build info
    with open(index) as f:
        source = f.read()
    product = _get_value(source, r"\s*productName:\s*[\"'](\w+)[\"']")
    company = _get_value(source, r"\s*companyName:\s*[\"'](\w+)[\"']")
    version = _get_value(source, r"\s*productVersion:\s*[\"']([\w\d\.]+)[\"']")
    # Get loader
    loader = next((file for file in path.glob("*/*.loader.js") if file.is_file()), None)
    if not loader:
        raise RuntimeError("Could not find WebGL loader script")
    # Get framework
    framework = next((file for file in path.glob("*/*.js") if file.is_file() and "." not in file.stem), None)
    if not framework:
        raise RuntimeError("Count not find WebGL framework script")
    # Get code
    code = next((file for file in path.glob("*/*.wasm") if file.is_file()), None)
    if not code:
        raise RuntimeError("Could not find WebGL WebAssembly binary")
    # Get data
    data = next((file for file in path.glob("*/*.data") if file.is_file()), None)
    if not data:
        raise RuntimeError("Could not find WebGL data file")
    # Upload
    key = uuid4().hex
    verbose = True
    loader = Storage.upload(loader, type=UploadType.Demo, name=loader.name, key=key, verbose=verbose)
    framework = Storage.upload(framework, type=UploadType.Demo, name=framework.name, key=key, verbose=verbose)
    data = Storage.upload(data, type=UploadType.Demo, name=data.name, key=key, verbose=verbose)
    code = Storage.upload(code, type=UploadType.Demo, name=code.name, key=key, verbose=verbose)
    # Create demo
    response = query(f"""
        mutation ($input: CreateDemoInput!) {{
            createDemo (input: $input) {{
                tag
                url
            }}
        }}
        """,
        {
            "input": {
                "tag": tag,
                "product": product,
                "company": company,
                "version": version,
                "loader": loader,
                "framework": framework,
                "code": code,
                "data": data
            }
        },
        access_key=get_access_key()
    )
    # Print
    demo = response["createDemo"]
    print_json(data=demo)

@app.command(name="open", help="Open a demo.")
def open_demo (
    tag: str=Argument(..., help="Demo tag.")
) -> None:
    response = query(f"""
        query ($input: DemoInput!) {{
            demo (input: $input) {{
                tag
                url
            }}
        }}
        """,
        { "input": { "tag": tag } }
    )
    demo = response["demo"]
    if demo:
        open_browser(demo["url"])
    else:
        print("[red italic]Demo was not found[/red italic]")

@app.command(name="delete", help="Delete a demo.")
def delete_demo (
    tag: str=Argument(..., help="Demo tag.")
) -> None:
    response = query(f"""
        mutation ($input: DeleteDemoInput!) {{
            deleteDemo (input: $input)
        }}
        """,
        { "input": { "tag": tag } },
        access_key=get_access_key()
    )
    result = response["deleteDemo"]
    print_json(data=result)

@app.callback()
def demo_options (
    learn: bool = Option(None, "--learn", callback=create_learn_callback("https://docs.natml.ai/graph/demos/type"), help="Learn about predictor demos in NatML.")
):
    pass

def _get_value (source: str, pattern: str) -> str:
    for line in source.splitlines():
        match = search(pattern, line)
        if match:
            return match[1]
    return None


# 
#   NatML
#   Copyright © 2023 NatML Inc. All Rights Reserved.
#

from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import List, Optional, Tuple, Union

from .api import query
from .dtype import Dtype
from .storage import Storage, UploadType
from .tag import Tag, parse_tag, serialize_tag

@dataclass
class Endpoint:
    """
    Predictor endpoint.

    Members:
        variant (str): Endpoint variant.
        url (str): Endpoint URL.
        type (EndpointType): Endpoint type.
        acceleration (EndpointAcceleration): Endpoint acceleration.
        status (EndpointStatus): Endpoint status.
        created (str): Date created.
        signature (EndpointSignature): Endpoint prediction signature. This is populated for `ACTIVE` endpoints.
        error (str): Endpoint error. This is populated for `INVALID` endpoints.
    """
    variant: str
    url: str
    type: EndpointType
    acceleration: EndpointAcceleration
    status: EndpointStatus
    created: str
    signature: Optional[EndpointSignature] = None
    error: Optional[str] = None
    FIELDS = f"""
    variant
    url
    type
    acceleration
    status
    created
    signature {{
        inputs {{
            name
            type
            description
            range
            optional
            string_default: stringDefault
            float_default: floatDefault
            int_default: intDefault
            bool_default: boolDefault
        }}
        outputs {{
            name
            type
            description
            range
            optional
            string_default: stringDefault
            float_default: floatDefault
            int_default: intDefault
            bool_default: boolDefault
        }}
    }}
    error
    """

    def __post_init__ (self):
        self.signature = EndpointSignature(**self.signature) if isinstance(self.signature, dict) else self.signature

    @classmethod
    def retrieve (
        cls,
        tag: str,
        access_key: str=None
    ) -> Endpoint:
        """
        Retrieve a predictor endpoint.

        Parameters:
            tag (str): Endpoint tag. If the tag does not contain a variant then the variant defaults to `main`.
            access_key (str): NatML access key.

        Returns:
            Endpoint: Predictor endpoint.
        """
        # Ensure this is a predictor tag
        tag = parse_tag(tag)
        variant = tag.variant or "main"
        tag = Tag(tag.username, tag.name)
        tag = serialize_tag(tag)
        # There isn't a query to get a specific endpoint so just filter for now
        endpoints = cls.list(tag, access_key=access_key)
        endpoint = next((x for x in endpoints if x.variant == variant), None) if endpoints else None
        # Return
        return endpoint

    @classmethod
    def list (
        cls,
        tag: str,
        access_key: str=None
    ) -> List[Endpoint]:
        """
        Retrieve all predictor endpoints.

        Parameters:
            tag (str): Predictor tag. This MUST NOT be a variant tag.
            access_key (str): NatML access key.

        Returns:
            list: Predictor endpoints.
        """
        # Request
        response = query(f"""
            query ($input: PredictorInput!) {{
                predictor (input: $input) {{
                    endpoints {{
                        {Endpoint.FIELDS}
                    }}
                }}
            }}""",
            { "input": { "tag": tag } },
            access_key=access_key
        )
        # Check predictor
        predictor = response["predictor"]
        if not predictor:
            return None
        # Get endpoints
        endpoints = [Endpoint(**endpoint) for endpoint in predictor["endpoints"]]
        return endpoints

    @classmethod
    def create (
        cls,
        tag: str,
        notebook: Union[str, Path],
        type: EndpointType,
        acceleration: EndpointAcceleration,
        environment: Optional[dict] = None,
        access_key: str=None
    ) -> Endpoint:
        """
        Create a predictor endpoint.

        Parameters:
            tag (str): Endpoint tag. If the tag does not contain a variant then the variant defaults to `main`.
            notebook (str | Path): Notebook URL or path.
            type (EndpointType): Endpoint type.
            acceleration (EndpointAcceleration): Endpoint acceleration.
            environment (dict): Environment variables.
            access_key (str): NatML access key.

        Returns:
            Endpoint: Created endpoint.
        """
        # Upload notebook
        url = Storage.upload(notebook, UploadType.Notebook, check_extension=True) if isinstance(notebook, Path) else notebook
        # Query
        environment = [{ "name": name, "value": value } for name, value in environment.items()] if environment is not None else None
        response = query(f"""
            mutation ($input: CreateEndpointInput!) {{
                createEndpoint (input: $input) {{
                    {Endpoint.FIELDS}
                }}
            }}""",
            { "input": { "tag": tag, "notebook": url, "type": type, "acceleration": acceleration, "environment": environment } },
            access_key=access_key
        )
        # Create endpoint
        endpoint = response["createEndpoint"]
        endpoint = Endpoint(**endpoint)
        # Return
        return endpoint

    @classmethod
    def delete (
        cls,
        tag: str,
        access_key: str=None
    ) -> bool:
        """
        Delete a predictor endpoint.

        Parameters:
            tag (str): Endpoint tag. If the tag does not contain a variant then the variant defaults to `main`.
            access_key (str): NatML access key.

        Returns:
            bool: Whether the endpoint was successfully deleted.
        """
        # Query
        response = query(f"""
            mutation ($input: DeleteEndpointInput!) {{
                deleteEndpoint (input: $input)
            }}""",
            { "input": { "tag": tag } },
            access_key=access_key
        )
        # Return
        result = response["deleteEndpoint"]
        return result
    
@dataclass
class EndpointSignature:
    """
    Endpoint signature.

    Members:
        inputs (list): Input parameters.
        outputs (list): Output parameters.
    """
    inputs: List[EndpointParameter]
    outputs: List[EndpointParameter]

    def __post_init__ (self):
        self.inputs = [EndpointParameter(**parameter) if isinstance(parameter, dict) else parameter for parameter in self.inputs]
        self.outputs = [EndpointParameter(**parameter) if isinstance(parameter, dict) else parameter for parameter in self.outputs]
    
@dataclass
class EndpointParameter:
    """
    Endpoint parameter.

    Members:
        name (str): Parameter name. This is only populated for input parameters.
        type (Dtype): Parameter type. This is `None` if the type is unknown or unsupported by NatML.
        description (str): Parameter description.
        optional (bool): Parameter is optional.
        range (tuple): Parameter value range for numeric parameters.
        stringDefault (str): Parameter default string value.
        floatDefault (float): Parameter default float value.
        intDefault (int): Parameter default integer value.
        boolDefault (bool): Parameter default boolean value.
    """
    name: Optional[str] = None
    type: Optional[Dtype] = None
    description: Optional[str] = None
    optional: Optional[bool] = None
    range: Optional[Tuple[float, float]] = None
    string_default: Optional[str] = None
    float_default: Optional[float] = None
    int_default: Optional[int] = None
    bool_default: Optional[bool] = None

class EndpointType (str, Enum):
    """
    Endpoint type.
    """
    Serverless = "SERVERLESS"
    Dedicated = "DEDICATED"

class EndpointAcceleration (str, Enum):
    """
    Endpoint acceleration.
    """
    CPU = "CPU"
    A40 = "A40"
    A100 = "A100"

class EndpointStatus (str, Enum):
    """
    Endpoint status.
    """
    Provisioning = "PROVISIONING"
    Active = "ACTIVE"
    Invalid = "INVALID"